#include	"stdThread.h"
#include	"SendRecvThread.h"	// 送受信スレッドを使うため
#include	"MySyncObject.h"	// 同期オブジェクトを使うため
#include	"define.h"

#ifdef _MSC_VER			// Windowsのとき
#define DISABLE_C4996   __pragma(warning(push)) __pragma(warning(disable:4996))
#define ENABLE_C4996    __pragma(warning(pop))
#else					// Linuxのとき
#define DISABLE_C4996
#define ENABLE_C4996
#endif

// メニューで使用する文字
#define		CMD_QUIT_CHAR		'q'
#define		CMD_SEND_MSG_CHAR	'm'

// 関数の宣言
BOOL CreateAndConnectSocket(LPCSTR szAddress, WORD wPort);	// ソケットの作成と接続処理
BOOL DestroySocket(SOCKET &fd);					// 切断とソケットの破棄
void Stop();									// すべてのソケットを破棄する
void DispMenu();								// メニューの表示
int	 GetKeyString(LPSTR pszString, int iSize);	// キーボード入力文字の取得
BOOL KillZombei();								// ゾンビ状態のスレッドを破棄する
BOOL ShowInterfaceIndex(addrinfo *pAddr);		// インターフェースを表示する
BOOL ShowPathMTU(SOCKET fd, int iFamily);		// MTUを表示する
BOOL SetHopLimit(SOCKET fd, int iFamily, int iLimit);	// ホップリミットを設定する
BOOL SendMessagePacket(char *pcData, int iSize);	// メッセージパケットの送信

// 変数の宣言
CSendRecvThread		*m_pCSendRecvThread = NULL;
ConnectionInfoRec	*m_pConInfo = NULL;
CMySyncObject		*m_pCMySyncObject = NULL;

int main(int argc, char *argv[])
{
	char	szKeyInBuff[81];

	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();

	// 起動パラメータチェック
	if (argc != 3)
	{
		fprintf(stderr, "Usage:  %s <ServerAddress> <ServerPort>\n", argv[0]);
		goto L_END;
	}
	// ソケットの作成と接続処理
	// 送受信スレッドを開始する
	if (CreateAndConnectSocket(argv[1], (WORD)atol(argv[2])) == FALSE)
		goto L_END;
	// メニューの表示
	DispMenu();
	while (1)
	{
		if (KillZombei() == TRUE)			// 切断してたら終了
			break;
		// キーボードから入力された文字列が'q'なら終了
		// 'm'ならメッセージを入力後は送信する		
		switch (GetKeyString(szKeyInBuff, sizeof(szKeyInBuff)))
		{
			case 0:							// 入力なし、何もしない
				break;
			case CMD_QUIT_CHAR:				// 'q'入力終了
				goto L_END;
			case CMD_SEND_MSG_CHAR:			// 'm'入力szKeyInBuffに文字列が格納されている
				// 送信は送信データをセットするだけ、実際の送信はCSendRecvThreadで実施
				// メッセージパケットの送信
				SendMessagePacket(szKeyInBuff, strlen(szKeyInBuff));
				// メニューの表示
				DispMenu();
				break;
		}
	}

L_END:
	// 切断とすべてのソケットの破棄
	Stop();
	m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	return(0);
}


//----------------------------------------------
// function
//  TCPソケットの作成, connectの実行
//  接続成功したとき送受信スレッドを開始する
// parameter
//  LPCSTR       szAddress      [in]接続先アドレス
//  WORD         wPort          [in]接続先ポート
// return
//  TRUE/FALSE
//----------------------------------------------
BOOL CreateAndConnectSocket(LPCSTR szAddress, WORD wPort)
{
	BOOL		fRet = TRUE;
	SOCKET      fd;
	struct addrinfo		hints, *pres = NULL, *pTemp = NULL;
	char		szPort[NI_MAXSERV];

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;		// TCP
	DISABLE_C4996
	sprintf(szPort, "%d", wPort);			// ポート番号=サービス
	ENABLE_C4996
	// アドレスを指定してgetaddrinfoを実行 接続先アドレス情報を得る
	if (getaddrinfo(szAddress, szPort, &hints, &pres) == 0)
	{
		pTemp = pres;
		while (pTemp != NULL)
		{
			// socket, connectを実行
			if ((fd = socket(pTemp->ai_family, pTemp->ai_socktype, pTemp->ai_protocol)) == INVALID_SOCKET)
				goto L_NEXT;
			// 使用するインターフェースを表示する
			ShowInterfaceIndex(pTemp);
			// HopLimitを取得設定する
			SetHopLimit(fd, pTemp->ai_family, 10);

			if (connect(fd, pTemp->ai_addr, (int)pTemp->ai_addrlen) == SOCKET_ERROR)
			{
				DispErrorMsg("Err:connect");
				DestroySocket(fd);
				goto L_NEXT;
			}
			// 接続が成功したのでMTUを調べる
			ShowPathMTU(fd, pTemp->ai_family);
			// 接続成功したのでこのソケット採用することにして抜ける
			fRet = TRUE;
			break;
L_NEXT:
			pTemp = pTemp->ai_next;
		}
		freeaddrinfo(pres);
	}
	if (fd == INVALID_SOCKET)	// 接続に失敗
	{
		DispErrorMsg("Err:connect");
		fRet = FALSE;
		goto L_END;
	}
	// 送信スレッドを開始する
	m_pConInfo = (ConnectionInfoRec *)calloc(1, sizeof(ConnectionInfoRec));
	m_pConInfo->fdClient = fd;
	m_pConInfo->pCMySyncObject = m_pCMySyncObject;
	m_pCSendRecvThread = new CSendRecvThread(m_pConInfo);
	m_pCSendRecvThread->Begin();

L_END:
	return(fRet);
}


//----------------------------------------------
// function
//    TCPソケットの破棄
// parameter
//    SOCKET		&fd[in/out]破棄するソケット
// return
//    TRUE/FALSE
//----------------------------------------------
BOOL DestroySocket(SOCKET &fd)
{
	if (fd != INVALID_SOCKET)
	{
		fprintf(stderr, "DestroySocket()\n");
		shutdown(fd, SHUT_RDWR);     // 受信も送信も停止
		close(fd);
		fd = INVALID_SOCKET;
	}
	return(TRUE);
}

//----------------------------------------------
// function
//    送受信スレッドの終了、ソケットの破棄
// parameter
//    なし
// return
//    なし
//----------------------------------------------
void Stop()
{
	fprintf(stderr, "Stop()\n");
	if (m_pCSendRecvThread != NULL)
	{
		m_pCSendRecvThread->End();
		m_pCSendRecvThread->WaitForEnd();
		SAFE_DELETE(m_pCSendRecvThread)

		DestroySocket(m_pConInfo->fdClient);
		SAFE_FREE(m_pConInfo)
	}
}

//----------------------------------------------
// function
//    メニューの表示
// parameter
//    なし
// return
//    なし
//----------------------------------------------
void DispMenu()
{
	fprintf(stderr, "\x1b[6;1H\x1b[0J");
	fprintf(stderr, "%c:quit %c:メッセージ送信 : ", CMD_QUIT_CHAR, CMD_SEND_MSG_CHAR);
}

#if 0
//----------------------------------------------
// function
//    キーボード入力(改行まで)の取得
//		入力がないときはすぐにリターン
// parameter
//    LPSTR		pszString	[in/out]データ格納用バッファ
//    int		iSize		[in]バッファのサイズ
// return
//    0:入力なし CMD_QUIT_CHAR:終了 CMD_SEND_MSG_CHAR:メッセージ送信
//----------------------------------------------
int GetKeyString(LPSTR pszString, int iSize)
{
	BOOL			fRet = FALSE;
	int				iRet = 0;

	if (kbhit())
	{
		switch (getchar())
		{
			case CMD_QUIT_CHAR:
				iRet = CMD_QUIT_CHAR;
				break;
			case CMD_SEND_MSG_CHAR:
				fprintf(stderr, "\nInput Msg : ");
				fgets(pszString, iSize, stdin);
				if (strlen(pszString) > 0)
				{
				// 改行まで読み込んでいるので
					if (pszString[strlen(pszString) - 1] == '\n')
						pszString[strlen(pszString) - 1] = 0;
					iRet = CMD_SEND_MSG_CHAR;
				}
				break;
			default:
				break;
		}
	}
	return(iRet);
}
#else
// リモートデバッグではttyじゃないのでkbhitは使えない
// 標準入力は0なのでselectのままでOKですが、練習のためにpollに
int GetKeyString(LPSTR pszString, int iSize)
{
	BOOL			fRet = FALSE;
	int				iRet = 0;
	pollfd			fds[1] = {0};

	fds[0].fd = 0;					// 0:標準入力 1:標準出力 2:標準エラー
	fds[0].events = POLLIN;			// Enterキー入力を対象に
	poll(fds, 1, 10);
	if (fds[0].revents & POLLIN)	// ENTERキー入力があった
	{
		memset(pszString, 0, iSize);
		read(0, pszString, iSize);
		switch (pszString[0])
		{
			case CMD_QUIT_CHAR:
				iRet = CMD_QUIT_CHAR;
				break;
			case CMD_SEND_MSG_CHAR:
				while (1)
				{
					poll(fds, 1, 10);
					if (fds[0].revents & POLLIN)
						read(0, pszString, iSize);
					else
						break;
				}
				fprintf(stderr, "\n送信メッセージを入力してください : ");
				memset(pszString, 0, iSize);
				read(0, pszString, iSize);	// ENTERキーまで読み込む
				if (strlen(pszString) > 0)
				{
				// 改行まで読み込んでいるので
					if (pszString[strlen(pszString) - 1] == '\n')
						pszString[strlen(pszString) - 1] = 0;
					iRet = CMD_SEND_MSG_CHAR;
				}
				break;
			default:
				break;
		}
	}
	return(iRet);
}
#endif

//----------------------------------------------
// function
//	切断済みのCSendRecvThreadを破棄する
// parameter
//	なし
// return
//	TRUE:破棄した/FALSE:破棄すべきものがなかった
//----------------------------------------------
BOOL KillZombei()
{
	BOOL	fRet = FALSE;

	if (m_pCSendRecvThread != NULL)
	{
		if (m_pCSendRecvThread->IsZombie() == TRUE)
		{
			m_pCSendRecvThread->End();
			m_pCSendRecvThread->WaitForEnd();
			SAFE_DELETE(m_pCSendRecvThread);

			DestroySocket(m_pConInfo->fdClient);
			SAFE_FREE(m_pConInfo)
			fRet = TRUE;
		}
	}
	return(fRet);
}

//----------------------------------------------------------------------
// function
//		接続に使用されるインターフェースインデックスを調べる
// parameter
//		addrinfo	*pAddr		[in]接続に使用されるアドレス情報
// return
//		TRUE/FALSE
//----------------------------------------------------------------------
BOOL ShowInterfaceIndex(addrinfo *pAddr)
{
	BOOL			fRet = FALSE;
	sockaddr_in6	addr6;

	if (pAddr->ai_family == AF_INET6)
	{
		memcpy(&addr6, pAddr->ai_addr, sizeof(sockaddr_in6));
		fprintf(stderr, "\x1b[1;1H\x1b[0J");
		fprintf(stderr, "Index:%d\n", addr6.sin6_scope_id);
		fRet = TRUE;
	}
	return(fRet);
}


//----------------------------------------------------------------------
// function
//		PATH_MTUを調べる
//		このオプションは処理系によって未対応なのでIPV6_PATHMTUの定義有無で判断する
//		通常はPATH_MTU=1500Bytesと想定しデータは1024Bytesにすれば問題ない
// parameter
//		SOCKET		fd		[in]接続済みソケット
//		int			iFamily	[in]アドレスファミリ(AF_INET/AF_INET6)
// return
//		TRUE/FALSE
//----------------------------------------------------------------------
BOOL ShowPathMTU(SOCKET fd, int iFamily)
{
	BOOL		fRet = FALSE;

#ifdef IPV6_PATHMTU
	if (iFamily == AF_INET6)
	{
		struct ip6_mtuinfo mtuinfo;
		socklen_t infolen = sizeof(mtuinfo);
		int iRet = getsockopt(fd, IPPROTO_IPV6, IPV6_PATHMTU, &mtuinfo, &infolen);
		if (iRet == 0)
		{
			fprintf(stderr, "\x1b[4;1H\x1b[0J");
			fprintf(stderr, "PATH MTU %d\n", mtuinfo.ip6m_mtu);
			fRet = TRUE;
		}
	}
#endif
	return(fRet);
}

//----------------------------------------------------------------------
// function
//		HopLimit(IPv6)/TTL(IPv4)を設定する
// parameter
//		SOCKET		fd			[in]接続済みソケット
//		int			iFamily		[in]アドレスファミリ(AF_INET/AF_INET6)
//		int			iHopLimit	[in](HopLimit - 1)台のルータを超えることができる
// return
//		TRUE/FALSE
//----------------------------------------------------------------------
BOOL SetHopLimit(SOCKET fd, int iFamily, int iLimit)
{
	BOOL		fRet = FALSE;
	int			iRet;
	DWORD		dwHoplimit;
	socklen_t	iOptlen;

	if (iFamily == AF_INET6)
	{
		iOptlen = sizeof(dwHoplimit);
		iRet = getsockopt(fd, IPPROTO_IPV6, IPV6_UNICAST_HOPS, (char *)&dwHoplimit, &iOptlen);
		if (iRet == 0)
		{
			fprintf(stderr, "\x1b[2;1H\x1b[0J");
			fprintf(stderr, "HopLimitOrg %d\n", dwHoplimit);
		}
		dwHoplimit = iLimit;
		iOptlen = sizeof(dwHoplimit);
		iRet = setsockopt(fd, IPPROTO_IPV6, IPV6_UNICAST_HOPS, (char *)&dwHoplimit, iOptlen);
		if (iRet == 0)
		{
			iRet = getsockopt(fd, IPPROTO_IPV6, IPV6_UNICAST_HOPS, (char *)&dwHoplimit, &iOptlen);
			if (iRet == 0)
			{
				fprintf(stderr, "\x1b[3;1H\x1b[0J");
				fprintf(stderr, "HopLimitNew %d\n", dwHoplimit);
				fRet = TRUE;
			}
		}
	}
	else if (iFamily == AF_INET)
	{
		iOptlen = sizeof(dwHoplimit);
		iRet = getsockopt(fd, IPPROTO_IP, IP_TTL, (char *)&dwHoplimit, &iOptlen);
		if (iRet == 0)
		{
			fprintf(stderr, "\x1b[2;1H\x1b[0J");
			fprintf(stderr, "TTL %d\n", dwHoplimit);
		}
		dwHoplimit = iLimit;
		iOptlen = sizeof(dwHoplimit);
		iRet = setsockopt(fd, IPPROTO_IP, IP_TTL, (char *)&dwHoplimit, iOptlen);
		if (iRet == 0)
		{
			iRet = getsockopt(fd, IPPROTO_IP, IP_TTL, (char *)&dwHoplimit, &iOptlen);
			if (iRet == 0)
			{
				fprintf(stderr, "\x1b[3;1H\x1b[0J");
				fprintf(stderr, "TTLNew %d\n", dwHoplimit);
				fRet = TRUE;
			}
		}
	}
	return(fRet);
}


//----------------------------------------------
// function
//		メッセージパケットの送信
// parameter
//		char	*pcData		[in]UTF-8データ(NULLターミネートなし)
//		int		iSize		[in]データサイズ
// retun
//		なし
//----------------------------------------------
BOOL	SendMessagePacket(char *pcData, int iSize)
{
	MsgDataRec	*pMsgData = NULL;
	int		iPacketSize = sizeof(HeaderRec) + iSize;		// パケットサイズ
	BOOL		fRet = FALSE;

	pMsgData = (MsgDataRec *)calloc(iPacketSize, sizeof(BYTE));	// パケット全体のエリアを確保

	memcpy(pMsgData->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pMsgData->header.wCommand = htons(CMD_MSG_DATA);
	pMsgData->header.wDataLen = htons(iSize);

	memcpy(pMsgData->bMsgData, pcData, iSize);
	fRet = m_pCSendRecvThread->SetSendData((char *)pMsgData, iPacketSize);

	SAFE_FREE(pMsgData)
	return(fRet);
}
